package edu.berkeley.qfat;
import java.awt.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import edu.berkeley.qfat.geom.*;
import edu.berkeley.qfat.geom.HasBindingGroup;
import edu.wlu.cs.levy.CG.KDTree;
import edu.berkeley.qfat.geom.Point;
import com.infomatiq.jsi.IntProcedure;

public class Mesh implements Iterable<Mesh.T> {

    public static final float EPSILON = (float)0.0001;
    public static final Random random = new Random();

    private RTree<T>         triangles = new RTree<T>();
    private PointSet<Vertex> vertices  = new PointSet<Vertex>();

    public boolean immutableVertices;
    public Mesh    error_against      = null;
    public double  error              = 0;

    public Mesh(boolean immutableVertices) { this.immutableVertices = immutableVertices; }

    public void makeVerticesImmutable() { this.immutableVertices = true; }
    public float error() { return (float)error; }

    public int size() { return vertices.size(); }
    public Iterable<Vertex> vertices() { return vertices; }
    public Iterator<T> iterator() { return triangles.iterator(); }

    public void rebindPoints() {
        // unbind all points
        for(Mesh.T t : this) {
            t.v1().unbind();
            t.v2().unbind();
            t.v3().unbind();
        }
        // ask edges to re-implement their bindings
        for(Mesh.T t : this) {
            t.e1().dobind();
            t.e2().dobind();
            t.e3().dobind();
        }
        System.out.println("rebound!");
    }

    public void transform(Matrix m) {
        ArrayList<Vertex> set = new ArrayList<Vertex>();
        for(Vertex v : vertices) set.add(v);
        for(Vertex v : set) v.transform(m.times(v.p), true, null);
    }

    public void rebuild() { /*vertices.rebuild();*/ }
    public Vec diagonal() { return vertices.diagonal(); }
    public Point centroid() { return vertices.centroid(); }
    public Vertex nearest(Point p) { return vertices.nearest(p); }

    /** compute the volume of the mesh */
    public float volume() {
        double total = 0;
        for(T t : this) {
            double area = t.area();
            Vec origin_to_centroid = new Vec(new Point(0, 0, 0), t.centroid());
            boolean facingAway = t.norm().dot(origin_to_centroid) > 0;
            double height = Math.abs(t.norm().dot(origin_to_centroid));
            total += ((facingAway ? 1 : -1) * area * height) / 3.0;
        }
        return (float)total;
    }


    // Vertexices //////////////////////////////////////////////////////////////////////////////

    /** a vertex in the mesh */
    public final class Vertex extends HasQuadric implements Visitor {
        public Point p, goodp;
        public Point oldp;
        E e;                // some edge *leaving* this point

        private boolean illegal = false;

        public Point getPoint() { return p; }
        public float error() { return olderror; }

        private Vertex(Point p) {
            this.p = p;
            this.goodp = p;
            if (vertices.get(p) != null) throw new Error();
            vertices.add(this);
        }

        public void reinsert() {
            vertices.remove(this);
            vertices.add(this);
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) e.t.reinsert();
        }

        public float olderror = 0;
        public void setError(float nerror) {
            error -= olderror;
            olderror = nerror;
            error += olderror;
        }

        public float averageTriangleArea() {
            int count = 0;
            float ret = 0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                ret += e.t.area();
                count++;
            }
            return ret/count;
        }
        public float averageEdgeLength() {
            int count = 0;
            float ret = 0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                ret += e.length();
                count++;
            }
            return ret/count;
        }

        public Matrix _recomputeFundamentalQuadric() {
            Matrix m = Matrix.ZERO;
            int count = 0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                m = m.plus(e.t.norm().fundamentalQuadric(e.t.centroid()));
                count++;
            }
            return m.times(1/(float)count);
        }

        public HasQuadric nearest() { return error_against==null ? null : error_against.vertices.nearest(p, this); }
        public void computeError() {
            if (error_against==null) return;
            float nerror =
                nearest_in_other_mesh != null
                ? nearest_in_other_mesh.fundamentalQuadric().preAndPostMultiply(p)
                : nearest().fundamentalQuadric().preAndPostMultiply(p);
            if (quadric_count != 0)
                nerror = (nerror + quadric.preAndPostMultiply(p))/(quadric_count+1);

            if (!immutableVertices && quadric_count == 0)
                nerror *= 2;
                
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                double ang = Math.abs(e.dihedralAngle());
                if (ang > Math.PI) throw new Error();
                float minangle = (float)(Math.PI * 0.8);
                if (ang > minangle) nerror += (ang - minangle);
                if (e.t.aspect() < 0.2) {
                    nerror += (0.2-e.t.aspect()) * 10;
                }
            }

            setError(nerror);
        }

        public boolean move(Matrix m, boolean ignoreProblems) {
            boolean good = true;

            //     t1' = M * t1
            //     t2' = t2.getMatrix(t1) * t1'
            //     t2' = t2.getMatrix(t1) * M * t1
            //     t1 =     t1.getMatrix(t2) * t2
            // M * t1 = M * t1.getMatrix(t2) * t2
            if (bindingGroup!=null && this != bindingGroup.getMaster()) {
                Matrix v = getBindingMatrix(bindingGroup.getMaster());
                return ((Vertex)bindingGroup.getMaster()).move(v.inverse().times(m).times(v), ignoreProblems);
            }

            if (bindingGroup != null) {
                Matrix m2 = null;
                for(int i=0; i<20 && !m.equals(m2); i++) {
                    m2 = m.times(bindingGroup.krank);
                    //System.out.println(m.minus(m2));
                }
                if (!m.equals(m2)) return true;
            }

            Point op = this.p;
            Point pt = m.times(this.p);
            for(Vertex v : (Iterable<Vertex>)getBoundPeers()) {
                Point pt2 = v.getBindingMatrix(this).times(pt);
                /*
                if (Math.abs( v.p.minus(pt2).mag() / pt.minus(op).mag() ) > 5)
                    throw new Error(v.p+" "+pt2+"\n"+op+" "+pt+"\n"+v.getBindingMatrix(this));
                if (Math.abs( v.p.minus(pt2).mag() / pt.minus(op).mag() ) < 1/5) throw new Error();
                */
                good &= v.transform(pt2,
                                    ignoreProblems, v.getBindingMatrix(this));
            }

            for(Vertex v : (Iterable<Vertex>)getBoundPeers())
                v.recomputeFundamentalQuadricIfNeighborChanged();
            return good;
        }

        /** does NOT update bound pairs! */
        private boolean transform(Point newp, boolean ignoreProblems, Matrix yes) {
            this.oldp = this.p;
            if (immutableVertices) throw new Error();

            unApplyQuadricToNeighbor();


            illegal = false;
            if (this.p.minus(newp).mag() > 0.1 && !ignoreProblems) {
            /*
                try {
                    throw new Exception(""+this.p.minus(newp).mag()+" "+ignoreProblems+" "+yes);
                } catch(Exception e) {
                    e.printStackTrace();
                }
            */
                illegal = true;
            }

            this.p = newp;
            reinsert();
            applyQuadricToNeighbor();

            if (!ignoreProblems) {
                checkLegality();
            }
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                e.p2.quadricStale = true;
            return !illegal;
        } 

        public void checkLegality() {
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                if (Math.abs(e.dihedralAngle()) > (Math.PI * 0.9) ||
                    Math.abs(e.next.dihedralAngle()) > (Math.PI * 0.9)) illegal = true;
                if (e.t.aspect() < 0.2) illegal = true;
            }
            if (!illegal) triangles.range(oldp, this.p, (Visitor<T>)this);
        }

        public void reComputeErrorAround() {
            reComputeError();
            if (nearest_in_other_mesh != null)
                nearest_in_other_mesh.reComputeError();
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                e.p2.reComputeError();
        }

        public boolean visit(Object o) {
            if (o instanceof Vertex)
                return ((Vertex)o).e != null && ((Vertex)o).norm().dot(Vertex.this.norm()) >= 0;
            T t = (T)o;
            if (illegal) return false;
            for(E e = Vertex.this.e; e!=null; e=e.pair.next==Vertex.this.e?null:e.pair.next) {
                if (!t.has(e.p1) && !t.has(e.p2) && e.intersects(t)) { illegal = true; }
                if (e.t != null) {
                    if (!e.t.has(t.e1().p1) && !e.t.has(t.e1().p2) && t.e1().intersects(e.t)) { illegal = true; }
                    if (!e.t.has(t.e2().p1) && !e.t.has(t.e2().p2) && t.e2().intersects(e.t)) { illegal = true; }
                    if (!e.t.has(t.e3().p1) && !e.t.has(t.e3().p2) && t.e3().intersects(e.t)) { illegal = true; }
                }
            }
            return !illegal;
        }

        public E getFreeIncident() {
            E ret = getFreeIncident(e, e);
            if (ret != null) return ret;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                System.out.println(e + " " + e.t);
            throw new Error("unable to find free incident to " + this);
        }

        public E getFreeIncident(E start, E before) {
            for(E e = start; e!=null; e=e.pair.next==before?null:e.pair.next)
                if (e.pair.p2 == this && e.pair.t == null && e.pair.next.t == null)
                    return e.pair;
            return null;
        }

        public E getE(Point p2) {
            Vertex v = vertices.get(p2);
            if (v==null) return null;
            return getE(v);
        }
        public E getE(Vertex p2) {
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                if (e.p1 == this && e.p2 == p2) return e;
            return null;
        }

        private void glNormal(GL gl) {
            Vec norm = norm();
            gl.glNormal3f(norm.x, norm.y, norm.z);
        }
        public Vec norm() {
            Vec norm = new Vec(0, 0, 0);
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                if (e.t != null)
                    norm = norm.plus(e.t.norm().times((float)e.prev.angle()));
            return norm.norm();
        }

        public void bindTo(Vertex p) { bindTo(Matrix.ONE, p); }
    }


    /** [UNIQUE] an edge */
    public final class E extends HasBindingGroup implements Comparable<E> {

        public final Vertex p1, p2;
        T t;     // triangle to our "left"
        E prev;  // previous half-edge
        E next;  // next half-edge
        E pair;  // partner half-edge
        boolean shattered = false;

        public boolean intersects(T t) { return t.intersects(p1.p, p2.p); }

        public void bindingGroupChanged(edu.berkeley.qfat.geom.BindingGroup newBindingGroup_) {

            edu.berkeley.qfat.geom.BindingGroup<E> newBindingGroup =
                (edu.berkeley.qfat.geom.BindingGroup<E>)newBindingGroup_;
            if (newBindingGroup==null) return;
            if (this==newBindingGroup.getMaster()) return;
            HashSet<E> nbg = new HashSet<E>();
            for(E eother : (Iterable<E>)newBindingGroup) nbg.add(eother);
            for(E eother : nbg) {
                if (next==null || prev==null) continue;
                if (eother.next==null || eother.prev==null) continue;
                if (next.isBoundTo(eother.pair.prev.pair) && !prev.isBoundTo(eother.pair.next.pair))
                    prev.bindTo(next.getBindingMatrix(eother.pair.prev.pair), eother.pair.next.pair);
                if (!next.isBoundTo(eother.pair.prev.pair) && prev.isBoundTo(eother.pair.next.pair))
                    next.bindTo(prev.getBindingMatrix(eother.pair.next.pair), eother.pair.prev.pair);

                if (next.isBoundTo(eother.next) && !prev.isBoundTo(eother.prev))
                    prev.bindTo(next.getBindingMatrix(eother.next), eother.prev);
                if (!next.isBoundTo(eother.next) && prev.isBoundTo(eother.prev))
                    next.bindTo(prev.getBindingMatrix(eother.prev), eother.next);
            }

        }

        public float stretchRatio() {
            Vertex nearest = error_against.nearest(midpoint());
            float nearest_distance = midpoint().distance(nearest.p);
            float other_distance =
                (p1.p.distance(error_against.nearest(p1.p).p)+
                 p2.p.distance(error_against.nearest(p2.p).p))/2;
            return nearest_distance/other_distance;
        }
        public float comparator() {
            return length();
            //return t==null?0:(1/t.aspect());
        }
        public int compareTo(E e) {
            return e.comparator() > comparator() ? 1 : -1;
        }
        public void bindEdge(E e, Matrix m) {
            _bindEdge(e, m);
            pair._bindEdge(e.pair, m);
        }
        public void _bindEdge(E e, Matrix m) {
            e = e.pair;
            /*
            //assumes edges are identical length at binding time
            Vec reflectionPlaneNormal = e.p2.p.minus(e.p1.p).norm();
            float a = reflectionPlaneNormal.x;
            float b = reflectionPlaneNormal.y;
            float c = reflectionPlaneNormal.z;
            Matrix reflectionMatrix =
                new Matrix( 1-2*a*a,  -2*a*b,  -2*a*c, 0,
                            -2*a*b,  1-2*b*b,  -2*b*c, 0,
                            -2*a*c,   -2*b*c, 1-2*c*c, 0,
                            0,       0,       0,       1);
            m = m.times(Matrix.translate(e.midpoint().minus(Point.ORIGIN))
                        .times(reflectionMatrix)
                        .times(Matrix.translate(Point.ORIGIN.minus(e.midpoint()))));
            System.out.println(reflectionPlaneNormal);
            System.out.println("  " + p1.p + " " + m.times(e.p1.p));
            System.out.println("  " + p2.p + " " + m.times(e.p2.p));
            */
            if (m.times(e.p1.p).minus(p1.p).mag() > EPSILON) throw new Error();
            if (m.times(e.p2.p).minus(p2.p).mag() > EPSILON) throw new Error();
            this.bindTo(m, e);
        }
        
        public void dobind() {
            for(E e : (Iterable<E>)getBoundPeers()) {
                if (e==this) continue;
                p1.bindTo(getBindingMatrix(e), e.p1);
                p2.bindTo(getBindingMatrix(e), e.p2);
            }
        }

        public Point shatter() {
            if (shattered || destroyed) return null;
            shattered = true;
            E first = null;
            E firste = null;
            E firstx = null;
            E firstq = null;
            for(E e : (Iterable<E>)getBoundPeers()) {
                E enext = e.next;
                E eprev = e.prev;
                E pnext = e.pair.next;
                E pprev = e.pair.prev;
                Point mid = e.midpoint();
                Vertex r = e.next.p2;
                Vertex l = e.pair.next.p2;
                e.destroy();
                e.pair.destroy();
                newT(r.p, e.p1.p, mid,    null, 0);
                newT(r.p, mid,    e.p2.p, null, 0);
                newT(l.p, mid,    e.p1.p, null, 0);
                newT(l.p, e.p2.p, mid,    null, 0);
            }
            for(E e : (Iterable<E>)getBoundPeers()) {
                Point mid = e.midpoint();
                if (first==null) {
                    first = e.p1.getE(mid);
                    firste = e;
                    firstx = e.pair;
                    firstq = e.p2.getE(mid).pair;
                    continue;
                }
                e.p1.getE(mid).          bindTo(e.getBindingMatrix(firste), first);
                e.p1.getE(mid).pair.     bindTo(e.getBindingMatrix(firste), first.pair);
                e.p2.getE(mid).pair.     bindTo(e.getBindingMatrix(firste), firstq);
                e.p2.getE(mid).pair.pair.bindTo(e.getBindingMatrix(firste), firstq.pair);
            }
            return null;
        }

        public boolean destroyed = false;
        public void destroy() {
            if (destroyed) return;
            destroyed = true;
            pair.destroyed = true;

            if (t != null) t.destroy();
            t = null;

            if (pair.t != null) pair.t.destroy();
            pair.t = null;

            if (next.t != null) next.t.destroy();
            if (prev.t != null) prev.t.destroy();
            next.t = null;
            prev.t = null;

            if (pair.next.t != null) pair.next.t.destroy();
            if (pair.prev.t != null) pair.next.t.destroy();
            pair.next.t = null;
            pair.prev.t = null;

            pair.prev.next = next;
            next.prev = pair.prev;
            prev.next = pair.next;
            pair.next = prev;
            if (p1.e == this) p1.e = prev.next;
            if (pair.p1.e == pair) pair.p1.e = pair.prev.next;
        }

        private void sync() {
            this.prev.next = this;
            this.next.prev = this;
            this.pair.pair = this;
            if (this.next.p1 != p2) throw new Error();
            if (this.prev.p2 != p1) throw new Error();
            if (this.p1.e == null) this.p1.e = this;
            if (!added) added = true;
        }
        private boolean added = false;

        public T makeT(int colorclass) { return t==null ? (t = new T(this, colorclass)) : t; }

        public double dihedralAngle() {
            Vec v1 = t.norm().times(-1);
            Vec v2 = pair.t.norm().times(-1);
            return Math.acos(v1.norm().dot(v2.norm()));
        }

        /** angle between this half-edge and the next */
        public double angle() {
            Vec v1 = next.p2.p.minus(p2.p);
            Vec v2 = this.p1.p.minus(p2.p);
            return Math.acos(v1.norm().dot(v2.norm()));
        }

        public void makeAdjacent(E e) {
            if (this.next == e) return;
            if (p2 != e.p1) throw new Error("cannot make adjacent -- no shared vertex");
            if (t != null || e.t != null) throw new Error("cannot make adjacent -- edges not both free ");

            E freeIncident = p2.getFreeIncident(e, this);

            e.prev.next = freeIncident.next;
            freeIncident.next.prev = e.prev;

            freeIncident.next = this.next;
            this.next.prev = freeIncident;
            
            this.next = e;
            e.prev = this;

            sync();
            freeIncident.sync();
        }

        /** creates an isolated edge out in the middle of space */
        public E(Point p1, Point p2) {
            if (vertices.get(p1) != null) throw new Error();
            if (vertices.get(p2) != null) throw new Error();
            this.p1 = new Vertex(p1);
            this.p2 = new Vertex(p2);
            this.prev = this.next = this.pair = new E(this, this, this);
            this.p1.e = this;
            this.p2.e = this.pair;
            sync();
        }

        /** adds a new half-edge from prev.p2 to p2 */
        public E(E prev, Point p) {
            Vertex p2;
            p2 = vertices.get(p);
            if (p2 == null) p2 = new Vertex(p);
            this.p1 = prev.p2;
            this.p2 = p2;
            this.prev = prev;
            if (p2.getE(p1) != null) throw new Error();
            if (p2.e==null) {
                this.next = this.pair = new E(this, this, prev.next);
            } else {
                E q = p2.getFreeIncident();
                this.next = q.next;
                this.next.prev = this;
                E z = prev.next;
                this.prev.next = this;
                this.pair = new E(q, this, z);
            }
            if (p2.e==null) p2.e = this.pair;
            sync();
        }

        /** adds a new half-edge to the mesh with a given predecessor, successor, and pair */
        public E(E prev, E pair, E next) {
            this.p1 = prev.p2;
            this.p2 = next.p1;
            this.prev = prev;
            this.next = next;
            this.pair = pair;
            sync();
        }
        public Point midpoint() { return new Point((p1.p.x+p2.p.x)/2, (p1.p.y+p2.p.y)/2, (p1.p.z+p2.p.z)/2); }
        public boolean has(Vertex v) { return v==p1 || v==p2; }
        public float length() { return p1.p.minus(p2.p).mag(); }
        public String toString() { return p1+"->"+p2; }

    }

    public E makeE(Point p1, Point p2) {
        Vertex v1 = vertices.get(p1);
        Vertex v2 = vertices.get(p2);
        if (v1 != null && v2 != null) {
            E e = v1.getE(v2);
            if (e != null) return e;
            e = v2.getE(v1);
            if (e != null) return e;
        }
        if (v1 != null) return new E(v1.getFreeIncident(), p2);
        if (v2 != null) return new E(v2.getFreeIncident(), p1).pair;
        return new E(p1, p2);
    }
    public boolean coalesce = false;
    private static float round(float f) {
        return Math.round(f*1000)/1000f;
    }
    public T newT(Point p1, Point p2, Point p3, Vec norm, int colorclass) {
        if (coalesce) {

            for(Vertex v : vertices) { if (p1.distance(v.p) < EPSILON) { p1 = v.p; break; } }
            for(Vertex v : vertices) { if (p2.distance(v.p) < EPSILON) { p2 = v.p; break; } }
            for(Vertex v : vertices) { if (p3.distance(v.p) < EPSILON) { p3 = v.p; break; } }
            /*
            p1 = new Point(round(p1.x), round(p1.y), round(p1.z));
            p2 = new Point(round(p2.x), round(p2.y), round(p2.z));
            p3 = new Point(round(p3.x), round(p3.y), round(p3.z));
            */
        }
        if (norm != null) {
            Vec norm2 = p3.minus(p1).cross(p2.minus(p1));
            float dot = norm.dot(norm2);
            //if (Math.abs(dot) < EPointSILON) throw new Error("dot products within evertsilon of each other: "+norm+" "+norm2);
            if (dot < 0) { Point p = p1; p1=p2; p2 = p; }
        }
        E e12 = makeE(p1, p2);
        E e23 = makeE(p2, p3);
        E e31 = makeE(p3, p1);
        while(e12.next != e23 || e23.next != e31 || e31.next != e12) {
            e12.makeAdjacent(e23);
            e23.makeAdjacent(e31);
            e31.makeAdjacent(e12);
        }
        T ret = e12.makeT(colorclass);
        if (e12.t == null) throw new Error();
        if (e23.t == null) throw new Error();
        if (e31.t == null) throw new Error();
        return ret;
    }

    /** [UNIQUE] a triangle (face) */
    public final class T extends Triangle {
        public final E e1;
        public final int color;
        public final int colorclass;

        T(E e1, int colorclass) {
            this.e1 = e1;
            E e2 = e1.next;
            E e3 = e2.next;
            if (e1==e2 || e1==e3) throw new Error();
            if (e3.next!=e1) throw new Error();
            if (e1.t!=null || e2.t!=null || e3.t!=null) throw new Error("non-manifold surface or disagreeing normals");
            e1.t = this;
            e1.next.t = this;
            e1.next.next.t = this;

            // FIXME: check for sealed/watertight surface once construction is complete (and infer normal(s)?)

            int color = Math.abs(random.nextInt());
            while(true) {
                color = color % 4;
                if (e1().pair.t != null && color == e1().pair.t.color) { color++; continue; }
                if (e2().pair.t != null && color == e2().pair.t.color) { color++; continue; }
                if (e3().pair.t != null && color == e3().pair.t.color) { color++; continue; }
                break;
            }
            this.color = color;
            this.colorclass = colorclass;
            triangles.add(this);
        }
        public E e1() { return e1; }
        public E e2() { return e1.next; }
        public E e3() { return e1.prev; }
        public Vertex v1() { return e1.p1; }
        public Vertex v2() { return e1.p2; }
        public Vertex v3() { return e1.next.p2; }
        public Point p1() { return e1.p1.p; }
        public Point p2() { return e1.p2.p; }
        public Point p3() { return e1.next.p2.p; }
        public boolean hasE(E e) { return e1==e || e1.next==e || e1.prev==e; }
        public boolean has(Vertex v) { return v1()==v || v2()==v || v3()==v; }

        public void removeFromRTree() { triangles.remove(this); }
        public void addToRTree() { triangles.insert(this); }
        public void destroy() { triangles.remove(this); }
        public void reinsert() { triangles.remove(this); triangles.add(this); }

        public boolean shouldBeDrawn() {

            if (e1().bindingGroupSize() <= 1) return false;
            if (e2().bindingGroupSize() <= 1) return false;
            if (e3().bindingGroupSize() <= 1) return false;

            return true;
        }

        /** issue gl.glVertex() for each of the triangle's points */
        public void glVertices(GL gl) {
            if (!shouldBeDrawn()) return;
            norm().glNormal(gl);
            Point p1 = v1().goodp;
            Point p2 = v2().goodp;
            Point p3 = v3().goodp;
            p1.glVertex(gl);
            p2.glVertex(gl);
            p3.glVertex(gl);
        }

    }
}
